// ESFrameModelSwarm.java

import swarm.Globals;
import swarm.Selector;
import swarm.defobj.Zone;
import swarm.defobj.SymbolImpl;

import swarm.activity.Activity;
import swarm.activity.ActionGroup;
import swarm.activity.ActionGroupImpl;
import swarm.activity.Schedule;
import swarm.activity.ScheduleImpl;

import swarm.objectbase.Swarm;
import swarm.objectbase.SwarmImpl;
import swarm.objectbase.VarProbe;
import swarm.objectbase.MessageProbe;
import swarm.objectbase.EmptyProbeMapImpl;

import swarm.collections.ListImpl;
import swarm.collections.ListIndex;
import swarm.collections.ArrayImpl;

import java.io.*;
import java.util.*;

/**
 * The Model contains the Units
 *
 * @author Pietro Terna, with contributions of Marco Remondino, Dario Landini, 
 * Nicola Ormezzano and Paolo Pelligra
 */
public class ESFrameModelSwarm extends SwarmImpl
{
    // simulation parameters

    /** the total number of operating Units we are using*/
    public int totalUnitNumber;
    /** the max number of steps to be done to complete an order */
    public int maxStepNumber;
    /** the max length of a step in a recipe (in units of time) */
    public int maxStepLength;
    /** the total production time used by done orders */
    public float totalProductionTime;
    /** the total length of the recipes of the done orders */
    public float totalRecipeLength;

    /** the order generator */
    public OrderGenerator orderGenerator;

    /** 
     * ActionGroup for holding an ordered sequence of action 
     * (the numbers of the model action names are the same used in the full
     * jES model, so some are lacking in this version)
    */
    public ActionGroup modelActions2, modelActions2b, 
	modelActions2generator;

    /** the Schedule operating in the Model*/
    public Schedule modelSchedule;

    /** the input class for unit parameters*/
    public UnitParameters unitParameters;

    /** the list of the operating units */
    public ListImpl unitList;
    /** its iterator */
    public ListIndex unitListIndex;

    /** the tool to assign order to units */
    AssigningTool assigningTool;

    /** the address of the probe, if we need to access it; remember casting
     *  (ESFrameModelSwarm$1$ESFrameModelProbeMap)
     */
    Object modelProbe;

    // the constructor of the model
    public ESFrameModelSwarm (Zone aZone) {
	super (aZone);

	// Now fill in various simulation parameters with default values.
	totalUnitNumber = 10;
	maxStepNumber=30;
	maxStepLength=1;

	// Build a customized probe map using a `local' subclass
	// (a special kind of Java `inner class') of the
	// EmptyProbeMapImpl class.
	// This building operation produces a separated class file, with
	// name  ESFrameModelSwarm$1$ESFrameModelProbeMap.class
	class ESFrameModelProbeMap extends EmptyProbeMapImpl {
	    public VarProbe probeVariable (String name) {
		return
		    Globals.env.probeLibrary.getProbeForVariable$inClass
		    (name, ESFrameModelSwarm.this.getClass ());
	    }
	    public MessageProbe probeMessage (String name) {
		return
		    Globals.env.probeLibrary.getProbeForMessage$inClass
		    (name, ESFrameModelSwarm.this.getClass ());
	    }
	    public void addVar (String name) {
		addProbe (probeVariable (name));
	    }
	    public void addMessage (String name) {
		addProbe (probeMessage (name));
	    }
	    public ESFrameModelProbeMap (Zone _aZone, Class aClass) {
		super (_aZone, aClass);
		addVar ("totalUnitNumber");
		addVar ("maxStepNumber");
		addVar ("maxStepLength");
    }
	}

	// Install our custom probeMap class directly into the
	// probeLibrary
	modelProbe=new ESFrameModelProbeMap (aZone, getClass ());
	Globals.env.probeLibrary.setProbeMap$For
	    ((ESFrameModelProbeMap) modelProbe, getClass ());
    }
    /** Build the model objects. */
    public Object buildObjects ()
    {
	int i;

	// allow our parent class to build anything
	super.buildObjects();

        // create the class to read acounting parameters
        unitParameters = new 
	    UnitParameters(getZone(), totalUnitNumber);
	unitParameters.readUnitData();

	// the list of the production units
	unitList = new ListImpl (getZone());
	// its iterator
	unitListIndex = unitList.listBegin(getZone());

	// the instance of assigningTool
	assigningTool = new AssigningTool (getZone(),
					   unitList);

	// the orderGenerator
	orderGenerator = new OrderGenerator
	    (getZone(), maxStepNumber, maxStepLength, unitList,
	     this, assigningTool);

	// create the units
	for (i = 1; i <= totalUnitNumber; i++)
	    {
		Unit aUnit;

		aUnit = new Unit  (getZone(), this, assigningTool);
		unitParameters.setParametersTo(i, aUnit);
		    
		// add the unit to the list
		unitList.addLast (aUnit);

	    }

	// creating the dictionary in orderGenerator
	orderGenerator.setDictionary();

	return this;
    }

    /**
     * Here is where the model schedule is built, the data structures
     * that define the simulation of time in the model. The core is an
     * actionGroup that has a list of actions. That's then put in a
     * Schedule.  */
    public Object buildActions ()
    {
	int i, ii;
	super.buildActions();

	modelActions2 = new ActionGroupImpl (getZone ());
	modelActions2generator = new ActionGroupImpl (getZone ());
	modelActions2b = new ActionGroupImpl (getZone ());

	/* producing */
	modelActions2.createActionForEach$message(unitList,
			         SwarmUtils.getSelector("Unit","unitStep1"));
	/* a new order; this step is placed here, after the production step,
	 * to align the diffusion of the order forms (orders under execution 
	 * - next step - or new ones - this step -) */
	modelActions2generator.createActionTo$message(orderGenerator,
			       SwarmUtils.getSelector(orderGenerator,
				     "createRandomOrderWithNSteps"));
	modelActions2b.createActionForEach$message(unitList,
			  SwarmUtils.getSelector("Unit","unitStep2"));

	// Then we create a schedule that executes the
	// modelActions. modelActions is an ActionGroup, by itself it
	// has no notion of time. In order to have it executed in
	// time, we create a Schedule that says to use the
	// modelActions ActionGroup at particular times
	modelSchedule = new ScheduleImpl (getZone (), 1);

	modelSchedule.at$createAction (0, modelActions2);
	modelSchedule.at$createAction (0, modelActions2generator);
	modelSchedule.at$createAction (0, modelActions2b);
	
	return this;
    }

    /**
     * Now set up the model's activation. swarmContext indicates where
     * we're being started in - typically, this model is run as a
     * subswarm of an observer swarm. */
    public Activity activateIn (Swarm swarmContext) {
	// First, activate ourselves via the superclass
	// activateIn: method.  Just pass along the context: the
	// activity library does the right thing.
	super.activateIn (swarmContext);

	// Now activate our own schedule.
	modelSchedule.activateIn (this);

	// Finally, return our activity.
	return getActivity ();
  }
    
    /** the method returns the list of the units */
    public Object getUnitList(){
	return unitList;
    }

    /** record total production time and total recipe length */
    public void setProductionTimeAndRecipeLength(float pt, float rl) {
	totalProductionTime += pt;
	totalRecipeLength   += rl;
	if(StartESFrame.verbose)
	    System.out.println("totalProductionTime " + totalProductionTime +
			       " totalRecipeLength " + totalRecipeLength +
			       " ratio " + ((float)
					    totalProductionTime)/totalRecipeLength);
	return;
    }

    /** report the ratio totalProductionTime/totalRecipeLength */
    public float getTimeLengthRatio() {
	if(totalRecipeLength != 0)
	    return totalProductionTime/totalRecipeLength;
	else return (float) 1; // 1 is a good choice, being 0 totalRecipeLength
	                       // in the first simulation days and being 1
	                       // the actual ratio of the first
	                       // concluded orders
    }

}


